/*
 * Created on 18.08.2003
 *
 */
package net.sourceforge.ganttproject.document;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.security.AccessControlException;

import org.apache.commons.httpclient.HttpException;
import org.apache.util.HttpURL;
import org.apache.webdav.lib.WebdavFile;
import org.apache.webdav.lib.WebdavResource;

/**
 * This class implements the interface Document for file access
 * on HTTP-servers and WebDAV-enabled-servers.
 * 
 * @author Michael Haeusler (michael at akatose.de)
 */
public class HttpDocument extends AbstractURLDocument {

	private String url;
	private String lastError;
	private HttpURL httpURL;
	private WebdavResource webdavResource;
	private WebdavFile webdavFile;
	private boolean locked = false;
	private boolean malformedURL = false;
	
	private static int lockDAVMinutes = 240;
	
	public HttpDocument (String url, String user, String pass) {
		this.url = url;
		httpURL = new HttpURL (url);
		
		try {
			if (user == null || user.length() == 0)
				user = httpURL.getUserName();
			if (pass == null || pass.length() == 0)
				pass = httpURL.getPassword();
			httpURL.setUserInfo(user, pass);
			WebdavResource.setGetUseDisk(false);
		} catch (MalformedURLException e) {
			lastError = e.getMessage();
			malformedURL = true;
		}
	}
	
	private WebdavResource getWebdavResource() {
		if (null == webdavResource)
			try {
				webdavResource = new WebdavResource(httpURL);
			} catch (HttpException e) {
				lastError = e.getMessage() + "(" + e.getStatusCode() + ")";
			} catch (IOException e) {
				lastError = e.getMessage();
			}
		return webdavResource;
	}
	
	private WebdavFile getWebdavFile() {
		if (null == webdavFile) {
			if (null != getWebdavResource())
				webdavFile = new WebdavFile(getWebdavResource());
		}
		return webdavFile;
	}
	
	public String getDescription() {
		try {
			return httpURL.getEscapedHttpURLExceptForUserInfo();
		} catch (MalformedURLException e) {
			return url;
		}
	}

	public boolean canRead() {
		return (null == getWebdavFile()) ? false : getWebdavFile().canRead();
	}

	public boolean canWrite() {
		if (null == getWebdavFile())
			return false;

		if (getWebdavFile().exists()) {
			return (getWebdavFile().canWrite());
		} else {
			return (true);
		}

/*		boolean writable = getWebdavFile().canWrite();
		if (! writable && ! getWebdavFile().exists()) {
			try {
				// TODO: find an alternative to test for the writability of a WebDAV location
				//       (the current implementation has locking problems on Tomcat)
				if (getWebdavFile().createNewFile()) {
					getWebdavFile().delete();
					writable = true;
				}
			} catch (IOException e) {
			}
		}
		return (writable); */
	}

	public boolean isValidForMRU() {
		return (! malformedURL);
	}

	/* (non-Javadoc)
	 * @see net.sourceforge.ganttproject.document.Document#acquireLock(java.lang.String)
	 */
	public boolean acquireLock() {
		if (locked) return true;
		if (null == getWebdavResource()) return false;
		try {
			String userName = " (GanttProject)";
			try {
				userName = " (" + System.getProperty("user.name") + "@GanttProject)";
			} catch (AccessControlException e) {
				// This can happen when running in a sandbox (Java WebStart)
				System.err.println (e + ": " + e.getMessage());
			}
			short lockSeconds = (short) ((lockDAVMinutes > 0) ? lockDAVMinutes * 60 : 10);
			locked = getWebdavResource().lockMethod(httpURL.getUserName() + userName, lockSeconds);
			return locked;
		} catch (HttpException e) {
			System.err.println(e.getMessage() + e.getStatusCode());
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println(e);
			e.printStackTrace();
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see net.sourceforge.ganttproject.document.Document#releaseLock()
	 */
	public void releaseLock() {
		if (null == getWebdavResource()) return;
		try {
			locked = false;
			getWebdavResource().unlockMethod();
		} catch (HttpException e) {
			System.err.println(e.getMessage() + e.getStatusCode());
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println(e);
			e.printStackTrace();
		}
	}

	public InputStream getInputStream() throws IOException {
		if (null == getWebdavResource())
			throw new IOException(lastError);
		try {
			return getWebdavResource().getMethodData();
		} catch (HttpException e) {
			throw new IOException(e.getMessage()
						  + "(" + e.getStatusCode() + ")");
		}
	}

	public OutputStream getOutputStream() throws IOException {
		if (null == getWebdavResource())
			throw new IOException(lastError);
		return new HttpDocumentOutputStream(getWebdavResource());
	}

	public String getPath() {
		return getDescription();
	}
	
	public String getURLPath() {
		return getPath();
	}

	public String getUsername() {
		try {
			return httpURL.getUserName();
		} catch (MalformedURLException e) {
			return null;
		}
	}

	public String getPassword() {
		try {
			return httpURL.getPassword();
		} catch (MalformedURLException e) {
			return null;
		}
	}

	public void setUserInfo(String user, String pass) {
		try {
			httpURL.setUserInfo(user, pass);

			webdavResource = null;
			webdavFile = null;
		} catch (MalformedURLException e) {
			lastError = e.getMessage();
			malformedURL = true;
		}
	}

	public String getLastError() {
		return lastError;
	}

	public static void setLockDAVMinutes(int i) {
		lockDAVMinutes = i;
	}
}
